#!/usr/bin/python
#
# Copyright 2006 Lars Wirzenius <liw@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import optparse
import stat
import sys
import time


from unitfmt import ByteFormatter, CountFormatter, TimeFormatter
from simpletask import BackgroundTask, TaskQueue


from summainlib import *


"""Check file integrity via checksums

See the manual page for more information.

Lars Wirzenius <liw@iki.fi>
"""


VERSION = "1.0"


class TerminalUserInterface:

    """User interface for the command line and text terminal"""
    
    def __init__(self, verbosity):
        self._verbosity = verbosity
        self._last = ""
        self._then = 0
        self._doingwhat = None
        self._pr = None
        self._output = sys.stdout
        self._sizefmt = None
        self._timefmt = TimeFormatter()
        self._task_start = None
        self._failures = 0
        self._successes = 0

    def _write(self, str):
        now = time.time()
        self._output.write("\b \b" * len(self._last))
        self._output.write(str)
        self._output.flush()
        self._last = str
        self._then = now
        
    def _clear(self):
        self._write("")
    
    def start_task(self, doingwhat, task, sizefmt):
        self._doingwhat = doingwhat
        self._pr = ProgressReporter(task)
        self._sizefmt = sizefmt
        self._task_start = time.time()

    def end_task(self):
        self._clear()
        if self._verbosity != "quiet":
            duration = time.time() - self._task_start
            (total, done, rt, speed) = self._pr.get_status()
            self._output.write("%s, total %s in %s, %s/s\n" % 
                               (self._doingwhat,
                                self._sizefmt.format(int(total), max=1),
                                self._timefmt.format(int(duration), max=2),
                                self._sizefmt.format(int(speed), max=1)))
        self._pr = None
        self._sizefmt = None
        self._task_start = None

    def report_misc_error(self, msg):
        self._clear()
        sys.stderr.write("ERROR: %s\n" % msg)
        self._failures += 1

    def report_oserror(self, filename, exception):
        self._clear()
        sys.stderr.write("ERROR: %s: %s\n" % (filename, exception.strerror))
        self._failures += 1

    def report_nonfile(self, filename):
        self._clear()
        sys.stderr.write("ERROR: %s is not a regular file\n" % filename)
        self._failures += 1

    def report_progress(self):
        if self._verbosity == "quiet":
            return
        (total, done, rt, speed) = self._pr.get_status()
        if total:
            percent = 100.0 * done / total
            if time.time() - self._then >= 1.0:
                total = self._sizefmt.format(total, max=1)
                done = self._sizefmt.format(done, max=1)
                rt = self._timefmt.format(int(rt), max=1)
                speed = self._sizefmt.format(int(speed), max=1)
                self._write(("%s, done %.1f %% at %s/s, %s remaining ") % 
                            (self._doingwhat, percent, speed, rt))
        else:
            if time.time() - self._then >= 1.0:
                done = self._sizefmt.format(done, max=1)
                speed = self._sizefmt.format(int(speed), max=1)
                self._write("%s, done %s; speed %s/s" % 
                            (self._doingwhat, done, speed))

    def report_success(self, filename):
        self._successes += 1
        if self._verbosity == "verbose":
            self._clear()
            self._output.write("OK: %s\n" % filename)

    def report_failure(self, filename):
        self._failures += 1
        self._clear()
        self._output.write("ERROR: %s\n" % filename)

    def report_final_results(self):
        if self._verbosity != "quiet":
            self._clear()
            if self._failures == 0:
                self._output.write("All files OK.\n")
            else:
                self._output.write("ERRORS: %d files bad, %d OK.\n" %
                                   (self._failures, self._successes))

    def get_failures(self):
        return self._failures


def parse_command_line():
    parser = optparse.OptionParser(version=VERSION)
    parser.add_option("-a", 
                      "--algorithm", 
                      action="append",
                      type="choice", 
                      choices=["md5", "sha1"],
                      dest="algorithms",
                      metavar="ALGORITHM",
                      help="Choose the algorith to use. Use multiple times " +
                           "to use several algorithms",
                      default=[])
    parser.add_option("-b", 
                      "--base", 
                      action="store",
                      dest="base",
                      default=None,
                      metavar="DIR",
                      help="Use DIR as the base directory for finding " +
                           "files. Filenames in the manifest file are " +
                           "relative to DIR. The default is the directory " +
                           "where the manifest file resides, or the " +
                           "current working directory if unknown.")
    parser.add_option("-c", 
                      "--check", 
                      action="store_true",
                      dest="check",
                      default=False,
                      help="Check files against an existing manifest. " +
                           "Default is to create a new manifest.")
    parser.add_option("-m", 
                      "--manifest-type", 
                      action="store",
                      type="choice", 
                      choices=["md5", "sha1", "dsc", "changes", "sources",
                               "signed"],
                      dest="manifest_type",
                      metavar="TYPE",
                      help="Choose the manifest file type.")
    parser.add_option("-o", 
                      "--output", 
                      action="store",
                      type="string", 
                      dest="output",
                      metavar="FILE",
                      help="Write the manifest into FILE. Default is stdout.")
    parser.add_option("-q", 
                      "--quiet", 
                      action="store_const",
                      dest="verbosity",
                      const="quiet",
                      default="terse",
                      help="Be quiet (do not report progress).")
    parser.add_option("-v", 
                      "--verbose", 
                      action="store_const",
                      dest="verbosity",
                      const="verbose",
                      help="Report status of every checked file, even if OK.")
    (options, filenames) = parser.parse_args()
    if not options.algorithms:
        options.algorithms = ["md5"]
    if not options.check and not options.manifest_type:
        options.manifest_type = options.algorithms[0]
    return options, filenames


def create_manifest(ui, options, filenames):
    if filenames:
        ff = FileFinder(filenames)
        ui.start_task("finding files", ff, CountFormatter())
        ff.start()
        while ff.do_work():
            ui.report_progress()
        ui.end_task()
        result = ff.get_result()
    else:
        result = [(None, None)]

    creator = ManifestCreator(ui, "", options.manifest_type,
                              options.algorithms,
                              result)
    ui.start_task("computing checksum", creator, ByteFormatter())
    creator.start()
    while creator.do_work():
        ui.report_progress()
    ui.end_task()
    if options.output:
        creator.get_manifest().write_to_named_file(options.output)
    else:
        creator.get_manifest().write(sys.stdout)


def check_manifests(ui, options, manifest_files):
    queue = TaskQueue()
    factory = ManifestParserFactory()
    guesser = ManifestTypeGuesser()

    if manifest_files:
        for filename in manifest_files:
            if options.base is None:
                base = os.path.dirname(filename)
            else:
                base = options.base
            f = file(filename, "r")
            data = f.read()
            f.close()
            if options.manifest_type:
                type = options.manifest_type
            else:
                type = guesser.guess(data)
                if type is None:
                    ui.report_misc_error("Can't deduce type of checksum " +
                                         "file %s" % filename)
                    return
            p = factory.get_parser(base, type)
            p.feed(data)
            p.close()
            checker = ManifestChecker(ui, p.get_manifest())
            queue.add(checker)
    else:
        if options.base is None:
            base = os.getcwd()
        else:
            base = options.base
        data = sys.stdin.read()
        if options.manifest_type:
            type = options.manifest_type
        else:
            type = guesser.guess(data)
            if type is None:
                ui.report_misc_error("Can't deduce type of checksum file " +
                                     "in stdin")
                return
        p = factory.get_parser(base, type)
        p.feed(data)
        p.close()
        checker = ManifestChecker(ui, p.get_manifest())
        queue.add(checker)

    ui.start_task("checking checksums", queue, ByteFormatter())
    queue.start()
    while queue.do_work():
        ui.report_progress()
    ui.end_task()
    ui.report_final_results()
    if ui.get_failures() > 0:
        sys.exit(1)
    
    
def main():
    (options, filenames) = parse_command_line()
    
    ui = TerminalUserInterface(options.verbosity)
    if options.check:
        check_manifests(ui, options, filenames)
    else:
        create_manifest(ui, options, filenames)


if __name__ == "__main__":
    main()
